#!/usr/bin/perl -w
# (c) Copyright 2005-2008. CodeWeavers, Inc.
package CXAssocGnome;
use strict;

use CXLog;
use CXUtils;
use CXAssoc;
use base "CXAssoc";


#####
#
# Package for reading and writing the Gnome .mime and .keys files.
#
#####

{
    package CXRWGnome;
    use strict;

    use CXLog;
    use CXUtils;

    sub new_blob($$$)
    {
        my ($self, $type, $line)=@_;
        my $blob={
            state => "up-to-date",
            type  => $type,
            index => scalar(@{$self->{blobs}}),
            lines => [ $line ]
        };
        push @{$self->{blobs}}, $blob;
        return $blob;
    }

    sub get($$)
    {
        my ($self, $section)=@_;
        my $blob=$self->{sections}->{$section};
        return undef if (!$blob);
        return undef if ($blob->{state} eq "deleted");
        return $blob;
    }

    sub get_created_by($$)
    {
        my ($self, $section)=@_;
        my $blob=$self->{created_by}->{$section};
        return undef if (!$blob);
        return undef if ($blob->{state} eq "deleted");
        return $blob;
    }

    sub modified($$)
    {
        my ($self, $blob)=@_;
        $blob->{state}="modified";
        $self->{modified}=1;
    }

    sub add($$$)
    {
        my ($self, $type, $section)=@_;
        my $blob;
        my $list=($type eq "created-by" ? "created_by" : "sections");
        $blob=$self->{$list}->{$section};
        if (!$blob)
        {
            $blob={
                type     => $type,
                section  => $section,
                state    => "modified"
            };
            $self->{$list}->{$section}=$blob;
            if ($type eq "created-by")
            {
                my $refblob=$self->{sections}->{$section};
                if ($refblob)
                {
                    # Add this comment next to the related section
                    my $blobs=$self->{blobs};
                    my $i=$refblob->{index};
                    while ($blobs->[$i])
                    {
                        if ($blobs->[$i] == $refblob)
                        {
                            $i++;
                            last;
                        }
                        $i++;
                    }
                    splice @$blobs, $i, 0, $blob;
                    $blob->{index}=$i;
                }
                else
                {
                    $blob->{index}=scalar(@{$self->{blobs}});
                    push @{$self->{blobs}}, $blob;
                }
            }
            else
            {
                $blob->{index}=scalar(@{$self->{blobs}});
                if ($blob->{index} > 0)
                {
                    my $last=$self->{blobs}->[-1];
                    if ($last->{type} ne "empty" or $last->{state} eq "deleted")
                    {
                        # Insert an empty line to keep things readable
                        $self->new_blob("empty", "");
                    }
                }
                push @{$self->{blobs}}, $blob;
            }
            $self->{modified}=1;
        }
        elsif ($blob->{state} eq "deleted")
        {
            # It's up to the caller to clean up this blob
            $blob->{state}="modified";
        }
        return $blob;
    }

    sub remove($$)
    {
        my ($self, $blob)=@_;

        # Never remove a blob, mark it as deleted
        $blob->{state}="deleted";

        # Also mark the previous blob as deleted if it is an empty blob
        my $blobs=$self->{blobs};
        my $i=$blob->{index};
        while ($blobs->[$i])
        {
            if ($blobs->[$i] == $blob)
            {
                if ($i > 0 and $blobs->[$i-1]->{type} eq "empty")
                {
                    cxlog("deleting empty blob\n");
                    $blobs->[$i-1]->{state}="deleted";
                }
                last;
            }
            $i++;
        }

        $self->{modified}=1;
    }

    sub remove_created_by($$)
    {
        my ($self, $blob)=@_;
        $blob->{state}="deleted";
        $self->{modified}=1;
    }

    sub new($$$)
    {
        my ($class, $filename, $delimiter)=@_;

        my $self={
            filename       => $filename,
            delimiter      => $delimiter,
            blobs          => [],    # Ordered list of the sections
            sections       => {}     # For quick access to sections
        };
        bless $self, $class;

        if (defined $filename and -e $filename)
        {
            my $fh;
            return undef if (!open($fh, "<", $filename));

            my $blob={ type => "raw" };
            while (my $line=<$fh>)
            {
                chomp $line;
                if ($line =~ /^\s*#\s*X-Created-By-([^= ]*)\s*=\s*(\S*)/)
                {

                    cxlog("createdby:[$line]\n");
                    my ($section, $apps)=($1, $2);
                    $section=demangle_string($section);
                    cxlog("section=[$section] apps=[$apps]\n");
                    $blob=$self->new_blob("created-by", $line);
                    $blob->{section}=$section;
                    map { $blob->{apps}->{$_}=1 } split /;+/, $apps;
                    $self->{created_by}->{$section}=$blob;
                }
                elsif ($line =~ /^\s*#/)
                {
                    if ($blob->{type} eq "comment")
                    {
                        cxlog("+:        [$line]\n");
                        push @{$blob->{lines}}, $line;
                    }
                    else
                    {
                        cxlog("comment:  [$line]\n");
                        $blob=$self->new_blob("comment", $line);
                    }
                }
                elsif ($line =~ /^\s*$/)
                {
                    if ($blob->{type} eq "empty")
                    {
                        cxlog("+:        [$line]\n");
                        push @{$blob->{lines}}, $line;
                    }
                    else
                    {
                        cxlog("empty:    [$line]\n");
                        $blob=$self->new_blob("empty", $line);
                    }
                }
                elsif ($blob->{type} eq "section" and
                       $line =~ /\s*(\S*)\s*$delimiter\s*(.*?)\s*$/)
                {
                    cxlog("field:    [$line]\n");
                    my ($field, $value)=($1, $2);
                    push @{$blob->{lines}}, $line;
                    # .mime files can have duplicate fields, but for these
                    # fields, all we care about are the 'ext' fields and
                    # they are taken care of below.
                    $blob->{fields}->{$field}=$value;
                    if ($field =~ /^ext$/i)
                    {
                        push @{$blob->{extlist}}, split / +/, $value;
                    }
                }
                elsif ($line =~ m!^(\S+)\s*$!)
                {
                    # A new blob
                    my $section=$1;
                    cxlog("section:  [$line]\n");
                    $blob=$self->new_blob("section", $line);
                    $blob->{section}=$section;
                    $blob->{fields}={};
                    $blob->{extlist}=[];
                    $self->{sections}->{$section}=$blob;
                }
                else
                {
                    # This is an unknown line format
                    $blob=$self->new_blob("raw", $line);
                    cxlog("unable to parse:\n", join("\n", @{$blob->{lines}}),
                          "\n");
                }
            }
            close($fh);
        }

        $self->{modified}=0;
        return $self;
    }

    sub isempty($)
    {
        my ($self)=@_;
        my $isempty=1;
        foreach my $blob (@{$self->{blobs}})
        {
            if ($blob->{state} ne "deleted")
            {
                $isempty=0;
                last;
            }
        }
        return $isempty;
    }

    sub save($)
    {
        my ($self)=@_;
        return 1 if (!$self->{modified});

        my $dir=cxdirname($self->{filename});
        if (!-d $dir and !cxmkpath($dir))
        {
            cxerr("unable to create the '$dir' directory: $@\n");
            return 0;
        }

        my $fh;
        if (!open($fh, ">", $self->{filename}))
        {
            cxerr("unable to open '$self->{filename}' for writing: $!\n");
            return 0;
        }

        cxlog("Saving '$self->{filename}'\n");
        foreach my $blob (@{$self->{blobs}})
        {
            if ($blob->{state} eq "up-to-date")
            {
                foreach my $line (@{$blob->{lines}})
                {
                    print $fh $line, "\n";
                }
            }
            elsif ($blob->{state} eq "modified")
            {
                # We have to rebuild the blob
                if ($blob->{type} eq "created-by")
                {
                    my $mangled=mangle_string($blob->{section});
                    print $fh "# X-Created-By-$mangled=",
                              join(";", keys %{$blob->{apps}}), "\n";
                }
                else
                {
                    if (defined $blob->{extlist} and @{$blob->{extlist}})
                    {
                        $blob->{fields}->{ext}=join(" ", @{$blob->{extlist}});
                    }
                    print $fh $blob->{section}, "\n";
                    my $delim=$self->{delimiter};
                    $delim=": " if ($delim eq ":");
                    foreach my $name (sort keys %{$blob->{fields}})
                    {
                        next if (!defined $blob->{fields}->{$name});
                        print $fh "\t$name$delim$blob->{fields}->{$name}\n";
                    }
                }
            }
        }
        close($fh);

        $self->{modified}=0;
        return 1;
    }

    sub delete_or_save($)
    {
        my ($self)=@_;
        if (!$self->isempty())
        {
            return $self->save();
        }
        if (-f $self->{filename})
        {
            cxlog("Deleting '$self->{filename}'\n");
            if (!unlink $self->{filename})
            {
                cxerr("unable to delete '$self->{filename}': $!\n");
                return 0;
            }
        }
        return 1;
    }
}


#####
#
# MIME database helper functions
#
#####

sub read_mime_db($)
{
    my ($self)=@_;
    return if ($self->{read_mime_db});
    $self->{read_mime_db}=1;

    foreach my $dir (split /:+/, $self->{mime_path})
    {
        my $dh;
        if (!opendir($dh, $dir))
        {
            cxlog("unable to open the '$dir' directory: $!\n");
            next;
        }
        foreach my $dentry (readdir $dh)
        {
            # Ignore CrossOver's MIME types
            next if ($dentry =~ /^cxassoc-/);

            # Ignore anything which is not a MIME type file
            next if ($dentry !~ /\.mime$/);
            $dentry="$dir/$dentry";
            next if (!-f $dentry);

            # Now we're left with the platform's native MIME types
            cxlog("Reading '$dentry'\n");
            my $rwmime=CXRWGnome->new($dentry, ":");
            while (my ($mimetype, $blob)=each %{$rwmime->{sections}})
            {
                # The extension lists of all MIME files are merged together
                $self->mdb_add_mime($mimetype, $blob->{extlist});
            }
        }
        closedir($dh);
    }
}


#####
#
# MIME icon creation and deletion
#
#####

sub create_icon($$)
{
    # Notes:
    # - SuSE 8.2 (Gnome 2.2.0) does not accept absolute paths or even
    #   filenames with an extension in the icon_filename field.
    #   What it wants is a base name with no extension, referring to either an
    #   XPM or PNG file in '<gnome-root>/pixmaps'. But this only for global
    #   icons, per-user icons seem broken. So we create tagged symbolic links
    #   pointing to the real icon files.
    # - On SuSE 8.2, putting the file in '<gnome-root>/icons' works too.
    # - Using the '<gnome-root>/pixmaps/gnome-mime-<mimetype>.<xpm|png>'
    #   convention would work too but generates more interference between
    #   bottles.
    # - RedHat 8.0 (Gnome 2.0.5) accepts absolute paths but not XPM files,
    #   both for global and per-user associations.

    my ($self, $mime)=@_;
    my $icon=$mime->{icon};
    $icon=~s/\.xpm$/.png/;
    $icon=$mime->{icon} if (!CXUtils::convert_to_png($mime->{icon}, $icon, 1));

    if ($self->{mime} eq $self->{global_mime})
    {
        $self->delete_icon($mime->{mimetype});
        my $base="cxassoc-$self->{tag}-" . mangle_string($mime->{mimetype});
        my $link=cxdirname($self->{mime}) . "/pixmaps/$base" .
                 ($icon =~ /\.png$/ ? ".png" : ".xpm");
        cxlog("Creating '$link'\n");
        return $base if (symlink $icon, $link);
        cxlog("unable to symlink '$link' to '$icon': $!\n");
    }

    return $icon;
}

sub delete_icon($$)
{
    my ($self, $mimetype)=@_;
    if ($self->{mime} eq $self->{global_mime})
    {
        my $base="cxassoc-$self->{tag}-" . mangle_string($mimetype);
        my $link=cxdirname($self->{mime}) . "/pixmaps/$base";
        foreach my $ext (".png", ".xpm")
        {
            if (-l "$link$ext")
            {
                cxlog("Deleting '$link$ext'\n");
                if (!unlink "$link$ext")
                {
                    cxlog("unable to delete '$link$ext': $!\n");
                }
            }
        }
    }
}

sub removeall_icon($$)
{
    my ($self, $pattern)=@_;
    if ($self->{mime} eq $self->{global_mime})
    {
        $pattern="^$pattern-.*\\.(?:png|xpm)\$";
        foreach my $dir (split /:+/, $self->{mime_path})
        {
            next if ($dir eq "");
            $dir=cxdirname($dir) . "/pixmaps";
            next if (!-d $dir);
            if (!-w $dir)
            {
                cxlog("skipping read-only '$dir' directory\n");
                next;
            }
            CXUtils::delete_files($dir, $pattern);
        }
    }
}


#####
#
# MIME type creation and deletion
#
#####

sub read_mime($)
{
    my ($self)=@_;
    if (!$self->{rwmime})
    {
        my $filename="$self->{mime}/cxassoc-$self->{tag}.mime";
        cxlog("Reading '$filename'\n");
        $self->{rwmime}=CXRWGnome->new($filename, ":");
    }
    return $self->{rwmime};
}

sub read_keys($)
{
    my ($self)=@_;
    if (!$self->{rwkeys})
    {
        my $filename="$self->{mime}/cxassoc-$self->{tag}.keys";
        cxlog("Reading '$filename'\n");
        $self->{rwkeys}=CXRWGnome->new($filename, "=");

        # Map the Gnome application ids to the keys sections
        while (my ($section, $cxkey)=each %{$self->{rwkeys}->{sections}})
        {
            foreach my $field ("default_application_id",
                               "short_list_application_user_additions")
            {
                my $value=$cxkey->{fields}->{$field} || "";
                # Note: short_list... is a comma separated list but not
                # default_application_id. Treat them the same anyway.
                foreach my $appid (split /,+/, $value)
                {
                    $self->{appid2keys}->{$appid}->{$section}=$cxkey;
                }
            }
        }
    }
    return $self->{rwkeys};
}

sub gnome_recode($$)
{
    my ($self, $str)=@_;
    # Notes:
    # - Gnome >= 2.0 expects UTF-8 files for the associations while
    #   older versions expect the files to be in the local encoding.
    # - Gnome < 2.0 also has problems if a string contains a '#'.
    if (!exists $self->{recode})
    {
        if ($self->{gnome_version} =~ /^1\./)
        {
            cxlog("Gnome 1.x detected\n");
            $self->{recode}=1;
            $self->{encoding}=CXUtils::get_system_encoding();
        }
        else
        {
            $self->{recode}=undef;
        }
    }
    if ($self->{recode})
    {
        $str =~ s/#//g;
        require CXRecode;
        $str=CXRecode::recode($str, "UTF-8", $self->{encoding});
    }
    return $str;
}

sub create_mime($$$$$$)
{
    my ($self, $domain, $massoc, $mime, $mimetype, $extensions)=@_;
    my $rwmime=$self->read_mime();

    # First the extensions in the .mime file
    my $cxmime=$rwmime->add("section", $mimetype);
    if (!$cxmime->{created})
    {
        cxlog("Creating .mime '$mimetype'\n");
        # Recreate the MIME type from scratch...
        $cxmime->{extlist}=$extensions;
        $cxmime->{fields}={};
        delete $cxmime->{lines};
        $cxmime->{created}=1;
        $rwmime->modified($cxmime);
    }
    else
    {
        cxlog("Tagging  .mime '$mimetype'\n");
    }
    my $created_by=$rwmime->add("created-by", $mimetype);
    if (!exists $created_by->{apps}->{$massoc->{id}})
    {
        $created_by->{apps}->{$massoc->{id}}=1;
        $rwmime->modified($created_by);
    }
    $self->{modified_mime}=1;

    # Then the description in the .keys file
    my $rwkeys=$self->read_keys();
    my $cxkey=$rwkeys->add("section", $mimetype);

    if (!$cxkey->{created})
    {
        cxlog("Creating .keys '$mimetype'\n");
        CXAssoc::setup_from_best_eassoc($mime);
        my $description=$mime->{description} || $mimetype;
        $description.=" (" . $self->id() . ")" if ($ENV{CX_TAGALL});

        # Preserve these two fields
        my $default=$cxkey->{fields}->{default_application_id};
        my $shortlist=$cxkey->{fields}->{short_list_application_user_additions};

        # And recreate the MIME type from scratch...
        $cxkey->{fields}={
            category      => "CrossOver",
            description   => $self->gnome_recode($description),
            icon_filename => $self->create_icon($mime)
        };
        if (defined $default or defined $shortlist)
        {
            $cxkey->{fields}->{default_action_type}="application";
            $cxkey->{fields}->{default_application_id}=$default;
            $cxkey->{fields}->{short_list_application_user_additions}=$shortlist;
        }
        delete $cxkey->{lines};
        $cxkey->{created}=1;
        $rwkeys->modified($cxkey);
    }
    else
    {
        cxlog("Tagging  .keys '$mimetype'\n");
    }
    $created_by=$rwkeys->add("created-by", $mimetype);
    if (!exists $created_by->{apps}->{$massoc->{id}})
    {
        $created_by->{apps}->{$massoc->{id}}=1;
        $rwkeys->modified($created_by);
    }

    return 1;
}

sub query_mime($$$$$)
{
    my ($self, $domain, $massoc, $mimetype, $extensions)=@_;

    my $rwmime=$self->read_mime();
    my $created_by=$rwmime->get_created_by($mimetype);
    return 0 if (!$created_by->{apps}->{$massoc->{id}});

    my $cxmime=$rwmime->get($mimetype);
    return 0 if (!$cxmime);
    if (!$cxmime->{extensions})
    {
        map { $cxmime->{extensions}->{$_}=1 } @{$cxmime->{extlist}};
    }
    foreach my $ext (@$extensions)
    {
        return 0 if (!$cxmime->{extensions}->{$ext});
    }

    # We don't check the description in the .keys file

    return 1;
}

sub untag_mime($$$$)
{
    my ($self, $domain, $massoc, $mimetype)=@_;

    # Untag the MIME type in the .mime file
    my $rwmime=$self->read_mime();
    my $created_by=$rwmime->get_created_by($mimetype);
    if ($created_by)
    {
        if ($created_by->{apps}->{$massoc->{id}})
        {
            delete $created_by->{apps}->{$massoc->{id}};
            $rwmime->modified($created_by);
        }

        if (!%{$created_by->{apps}})
        {
            # No one is using this MIME type anymore
            cxlog("Deleting mime - $mimetype\n");
            $rwmime->remove_created_by($created_by);
            my $cxmime=$rwmime->get($mimetype);
            $rwmime->remove($cxmime) if ($cxmime);
        }
    }

    # Then the MIME type in the .keys file
    my $in_use;
    my $rwkeys=$self->read_keys();
    $created_by=$rwkeys->get_created_by($mimetype);
    if ($created_by)
    {
        if ($created_by->{apps}->{$massoc->{id}})
        {
            delete $created_by->{apps}->{$massoc->{id}};
            $rwkeys->modified($created_by);
        }

        if (!%{$created_by->{apps}})
        {
            # No one is using this MIME type anymore
            cxlog("Deleting keys - $mimetype\n");
            $rwkeys->remove_created_by($created_by);
            my $cxkey=$rwkeys->get($mimetype);
            $rwkeys->remove($cxkey) if ($cxkey);
        }
        else
        {
            $in_use=1;
        }
    }
    $self->delete_icon($mimetype) if (!$in_use);

    return 1;
}


#####
#
# Association creation and deletion
#
#####

sub read_applications($)
{
    my ($self)=@_;
    if (!$self->{rwapps})
    {
        my $filename="$self->{application}/cxassoc-$self->{tag}.applications";
        cxlog("Reading '$filename'\n");
        $self->{rwapps}=CXRWGnome->new($filename, "=");
    }
    return $self->{rwapps};
}

sub gnome_app_id($$)
{
    my ($self, $massoc)=@_;
    my $app_id="$self->{tag}/$massoc->{id}";
    $app_id =~ s!:!/!g;
    return $app_id;
}

sub create_association($$$$)
{
    my ($self, $massoc, $adata, $extensions)=@_;
    my $rwapps=$self->read_applications();

    my $app_id=$self->gnome_app_id($massoc);
    cxlog("Creating '$app_id' association\n");
    my $cxassoc=$rwapps->add("section", $app_id);

    CXAssoc::setup_from_best_eassoc($massoc);
    my $name="";
    if ($massoc->{verb})
    {
        CXAssoc::compute_verb_name($massoc);
        if ($massoc->{verbname})
        {
            $name="$massoc->{verbname} with ";
            # The first '&' does not define a keyboard shortcut here
            $name =~ s/&//;
        }
    }
    $name.=$massoc->{appname};
    # Note: The combination of '.=', the above ampersand substitution,
    # UTF-8 accents, LANG=en_US.UTF-8 and CX_TAGALL=1 is causing
    # perl 5.8.0 to segfault on RedHat 8.0! So we avoid '.=' here.
    $name="$name (" . $self->id() . ")" if ($ENV{CX_TAGALL});

    # Note:
    # - Some versions of Gnome (2.4 on SUSE 9.1 and 2.6 on Mandrake 10.1)
    #   get confused if the first part of the command is a quoted path.
    #   So use 'sh --' to make them happy.
    # - 'sh --' works because we know the next parameter is a shell script
    #   (as created by CXAssocCheck) and not a binary executable.
    my $cmd="sh -- $massoc->{command}";

    # Recreate the association from scratch...
    $cxassoc->{fields}={
        name       => $self->gnome_recode($name),
        command    => $self->gnome_recode($cmd),
        mime_types => join(",", keys %{$adata->{all_mimes}})
    };
    delete $cxassoc->{lines};
    $rwapps->modified($cxassoc);

    return 1;
}

sub query_association($$$$)
{
    my ($self, $massoc, $adata, $state)=@_;
    my $rwapps=$self->read_applications();

    my $cxassoc=$rwapps->get($self->gnome_app_id($massoc));
    return $state if (!$cxassoc);

    # Note: The part we match will not have been recoded because it is mangled
    my $str=$cxassoc->{fields}->{command} || "";
    return $state if ($str !~ /\Q$massoc->{cmdbase}\E(?:[^:]|$)/);

    $str=$cxassoc->{fields}->{mime_types} || "";
    my @mimes=split /,+/, $str;
    return $state if (CXAssoc::compare_sets(\@mimes, $adata->{all_mimes}));

    return "alternative";
}

sub delete_association($$$)
{
    my ($self, $massoc, $adata)=@_;
    my $rwapps=$self->read_applications();

    my $app_id=$self->gnome_app_id($massoc);
    my $cxassoc=$rwapps->get($app_id);
    if ($cxassoc)
    {
        cxlog("Deleting the '$app_id' association\n");
        $rwapps->remove($cxassoc);
    }
    return 1;
}


#####
#
# Default creation and deletion
#
#####

sub create_default($$$$)
{
    my ($self, $massoc, $adata, $extensions)=@_;
    my $rwkeys=$self->read_keys();

    my $app_id=$self->gnome_app_id($massoc);
    foreach my $mimetype (keys %{$adata->{default_mimes}})
    {
        my $cxkey=$rwkeys->add("section", $mimetype);
        $cxkey->{fields}->{default_action_type}="application";
        $cxkey->{fields}->{default_application_id}=$app_id;
        my @shortlist=($app_id);
        my $shortlist=$cxkey->{fields}->{short_list_application_user_additions} || "";
        foreach my $item (split /,+/, $shortlist)
        {
            push @shortlist, $item if ($item ne $app_id);
        }
        $cxkey->{fields}->{short_list_application_user_additions}=join(",", @shortlist);
        $rwkeys->modified($cxkey);
        my $created_by=$rwkeys->add("created-by", $mimetype);
        if (!exists $created_by->{apps}->{$massoc->{id}})
        {
            $created_by->{apps}->{$massoc->{id}}=1;
            $rwkeys->modified($created_by);
        }
        $self->{appid2keys}->{$app_id}->{$mimetype}=$cxkey;
    }

    return 1;
}

sub query_default($$$$)
{
    my ($self, $massoc, $adata, $state)=@_;
    return "alternative" if (!%{$adata->{default_mimes}});
    my $rwkeys=$self->read_keys();

    my $app_id=$self->gnome_app_id($massoc);
    foreach my $mimetype (keys %{$adata->{default_mimes}})
    {
        my $cxkey=$rwkeys->get($mimetype);
        if (!$cxkey or
            $cxkey->{state} eq "deleted" or
            ($cxkey->{fields}->{default_action_type} || "") ne "application" or
            ($cxkey->{fields}->{default_application_id} || "") ne $app_id)
        {
            return $state;
        }
        my $str=$cxkey->{fields}->{short_list_application_user_additions};
        if (!$str or !grep /^$app_id$/, split /,+/, $str)
        {
            return $state;
        }
    }
    return "default";
}

sub delete_default($$)
{
    my ($self, $massoc)=@_;
    my $rwkeys=$self->read_keys();
    cxlog("delete_default $massoc->{id}\n");

    my $app_id=$self->gnome_app_id($massoc);
    while (my ($section, $cxkey)=each %{$self->{appid2keys}->{$app_id}})
    {
        next if ($cxkey->{state} eq "deleted");
        my $modified;

        my $str=$cxkey->{fields}->{default_application_id} || "";
        if ($str eq $app_id)
        {
            delete $cxkey->{fields}->{default_application_id};
            $modified=1;
        }

        $str=$cxkey->{fields}->{short_list_application_user_additions} || "";
        my $nochange=1;
        my @list=grep { $_ eq $app_id ? $nochange=0 : 1 } split /,+/, $str;
        if (!$nochange)
        {
            if (@list)
            {
                $cxkey->{fields}->{short_list_application_user_additions}=join(",", @list);
            }
            else
            {
                delete $cxkey->{fields}->{short_list_application_user_additions};
            }
            $modified=1;
        }

        if ($modified)
        {
            delete $cxkey->{fields}->{default_action_type};
            if (%{$cxkey->{fields}})
            {
                cxlog("Undefaulted keys - $section\n");
                $rwkeys->modified($cxkey);
            }
            else
            {
                cxlog("Deleting keys - $section\n");
                my $created_by=$rwkeys->get_created_by($section);
                $rwkeys->remove_created_by($created_by) if ($created_by);
                $rwkeys->remove($cxkey);
            }
        }
    }
    delete $self->{appid2keys}->{$app_id};
    return 1;
}


#####
#
# Main
#
#####

sub detect($$$$)
{
    my ($class, $cxoptions, $cxconfig, $gui_info)=@_;
    return () if (!$gui_info->{gnome_on} or
                  !$gui_info->{gnome_preferred_mime} or
                  !$gui_info->{gnome_preferred_application});

    my $self={
        gnome_version  => $gui_info->{gnome_version},
        application    => $gui_info->{gnome_preferred_application},
        mime           => $gui_info->{gnome_preferred_mime},
        mime_path      => $gui_info->{gnome_mime_path},
        global_mime    => $gui_info->{gnome_global_mime},
        do_assoc       => 1,
        do_default     => 1
    };
    bless $self, $class;
    $self->init_mime_handler($cxoptions);
    return ($self);
}

sub id($)
{
    my ($self)=@_;
    my $id="CXAssocGnome/$self->{application}";
    $id =~ s%/+%/%g;
    return $id;
}

sub preinstall($$)
{
    my ($self, $massoc)=@_;
    return $self->collect_unix_extensions($massoc);
}

sub install($$)
{
    my ($self, $massoc)=@_;
    return $self->action($self, $massoc, "install");
}

sub query($$)
{
    my ($self, $massoc)=@_;

    if (!$massoc)
    {
        my $id=$self->id();
        return {default     => $id,
                alternative => $id,
                mime        => $id,
                partial     => $id};
    }
    return $self->action($self, $massoc, "query");
}

sub uninstall($$)
{
    my ($self, $massoc)=@_;
    return $self->action($self, $massoc, "uninstall");
}

sub removeall($$)
{
    my ($self, $pattern)=@_;
    if ($pattern eq "legacy")
    {
        # FIXME: Some old versions of CrossOver used user.applications
        #        and user.keys.
        # Note: Some old versions of CrossOver used user.mime but this file
        # cannot be cleaned up because nothing distinguished the CrossOver
        # MIME types from the normal ones.
        my $productid=CXUtils::get_product_id();
        $pattern="^$productid\\.(?:mime|keys)\$";
        foreach my $dir (split /:+/, $self->{mime_path})
        {
            next if (!-d $dir);
            if (!-w $dir)
            {
                cxlog("skipping read-only '$dir' directory\n");
                next;
            }
            CXUtils::delete_files($dir, $pattern);
        }

        my $filename="$self->{application}/$productid.applications";
        if (-f $filename)
        {
            cxlog("Deleting '$filename'\n");
            if (!unlink $filename)
            {
                cxerr("unable to delete '$filename': $!\n");
            }
        }
        return 1;
    }

    $pattern="$pattern.*" if ($pattern !~ s/\$$//);
    $pattern="^cxassoc-$pattern";
    my $pat="$pattern\\.(?:mime|keys)\$";
    foreach my $dir (split /:+/, $self->{mime_path})
    {
        next if (!-d $dir);
        if (!-w $dir)
        {
            cxlog("skipping read-only '$dir' directory\n");
            next;
        }
        CXUtils::delete_files($dir, $pat);
    }
    $self->removeall_icon($pattern);

    CXUtils::delete_files($self->{application}, "$pattern\\.applications\$");

    return 1;
}

sub finalize($)
{
    my ($self)=@_;
    my $rc=1;
    $rc&=$self->{rwmime}->delete_or_save() if (defined $self->{rwmime});
    $rc&=$self->{rwkeys}->delete_or_save() if (defined $self->{rwkeys});
    $rc&=$self->{rwapps}->delete_or_save() if (defined $self->{rwapps});
    return $rc;
}

return 1;
