#!/usr/bin/perl -w
# (c) Copyright 2002-2008. CodeWeavers, Inc.
package CXUtils;
use strict;

# Define the module interface
use vars qw(@ISA @EXPORT);
use Exporter ();
@ISA    = "Exporter";
@EXPORT = qw(cxbackquote cxexec cxgettext cxmessage cxdirname cxbasename cxmkpath cxsystem demangle_string escape_string expand_string mangle_string mangle_path shquote_string unescape_string);

use CXLog;


# Portable which(1) implementation
sub cxwhich($$;$)
{
    my ($dirs, $app, $noexec)=@_;
    if ($app =~ /^\//)
    {
        return $app if ((-x $app or $noexec) and -f $app);
    }
    elsif ($app =~ /\//)
    {
        require Cwd;
        my $path=Cwd::cwd() . "/$app";
        return $path if ((-x $path or $noexec) and -f $path);
    }
    else
    {
        foreach my $dir (split /:/, $dirs)
        {
            return "$dir/$app" if ($dir ne "" and (-x "$dir/$app" or $noexec) and -f "$dir/$app");
        }
    }
    return undef;
}

# Find a terminal emulator
sub get_terminal_emulator(;$)
{
    my ($title)=@_;

    my $path;
    foreach my $term (qw(x-terminal-emulator xterm rxvt-xterm rxvt.bin rxvt eterm Eterm gnome-terminal konsole dtterm kvt aterm))
    {
        my $term_path=cxwhich("$ENV{PATH}:/usr/openwin/bin",$term);
        if (defined $term_path and -s $term_path)
        {
            $path=$term_path;
            last;
        }
    }
    return undef if (!$path);
    return ($path) if (!defined $title or $path =~ m%/x-terminal-emulator$%);
    return ($path,"-t",$title) if ($path =~ m%/gnome-terminal$%);
    return ($path,"-T",$title);
}



#####
#
# Product information
#
#####

sub check_product_id($)
{
    my ($productid)=@_;

    if (length($productid) < 4)
    {
        return "product id '$productid' is too short";
    }
    if ($productid !~ /^\w+$/)
    {
        return "invalid product id '$productid'";
    }
    return undef;
}

sub get_real_product_id()
{
    return "cxgames";
}

my $productid;
sub get_product_id(;\$)
{
    if (!defined $productid)
    {
        if (-f "$ENV{CX_ROOT}/.productid")
        {
            if (open(my $fh, "<", "$ENV{CX_ROOT}/.productid"))
            {
                $productid=<$fh>;
                chomp $productid;
                close($fh);
            }
        }
        else
        {
            $productid=get_real_product_id();
        }
        my $err=check_product_id($productid || "");
        if ($err)
        {
            if (@_ >= 1)
            {
                $_[0]=$err;
                return undef;
            }
            cxerr("$err\n");
            exit 1;
        }
    }
    return $productid;
}

sub get_product_name()
{
    return "CrossOver Games";
}

sub get_product_version()
{
    return "7.1.0";
}

sub get_std_icon($)
{
    my ($basename)=@_;
    return "$ENV{CX_ROOT}/share/icons/$basename";
}


#####
#
# String functions
#
#####

sub hash_string($)
{
    my ($str)=@_;

    use integer;
    my $hash=0;
    foreach my $char (split //, $str)
    {
        $hash = $hash*33 + ord($char);
        $hash = $hash + ($hash >> 5);
    }
    return $hash;
}

sub base32($$)
{
    my ($value, $digits)=@_;

    my $str="";
    my $base32="0123456789abcdefghijklmnopqrstuv";
    for (my $i=0; $i < $digits; $i++)
    {
        $str=substr($base32, ($value & 0x1f), 1) . $str;
        $value=($value >> 5);
    }
    return $str;
}

sub expand_tilde($)
{
    my ($user)=@_;
    my $subst;

    if (!$user)
    {
        $subst=$ENV{HOME} || (getpwuid($>))[7];
    }
    else
    {
        $subst=(getpwnam($user))[7];
    }
    return $subst || "~$user";
}

sub expand_string($)
{
    my ($str)=@_;
    $str =~ s+^~([^/]*)/+expand_tilde($1) . "/"+e;
    $str =~ s+:~([^/]*)/+":" . expand_tilde($1) . "/"+e;
    $str=~ s+\$\{([^}]*)\}+$ENV{$1} || ""+ge;
    return $str;
}

sub escape_string($)
{
    my ($str)=@_;
    $str =~ s%\\%\\\\%g;
    $str =~ s%\"%\\\"%g;
    return $str;
}

sub unescape_string($)
{
    my ($str)=@_;
    $str =~ s%\\\"%\"%g;
    $str =~ s%\\\\%\\%g;
    return $str;
}

# Quotes strings so they can be used in shell commands
# Note that this implies escaping '$'s and '`'s which may not be appropriate
# in another context.
sub shquote_string($)
{
    my ($str)=@_;
    $str =~ s%\\%\\\\%g;
    $str =~ s%\$%\\\$%g;
    $str =~ s%\"%\\\"%g;
    $str =~ s%\`%\\\`%g;
    return "\"$str\"";
}

# Mangles a string so that the result is:
# - shell safe: the result does not contain characters that may be interpreted
#   by the shell: $ " ' ` < > #
# - path safe: the result does not contain characters that would be a problem
#   in a file path: / \, but also spaces and accentuated characters to avoid
#   problems with different encodings across systems
# - XML safe: the result can be used as is in XML, that is the < > and &
#   characters are removed
# - list safe: the result can be put in a list which is either colon or
#   semi-colon separated
# - KDE Exec safe: the % character is encoded and no % character is introduced
#
# The mangling operation is 100% reversible. The mangling algorithm also
# strives to keep the result as readable as possible so it uses the following
# special conversions:
#   '/' -> '_'
#   '\' -> '~'
#   ' ' -> '+'
# All other troublesome characters, are encoded by converting them to
# hexadecimal and preceding them with a '^'.
sub mangle_string($)
{
    my ($str)=@_;
    $str =~ s!([\x01-\x1f\x80-\xff\$"'`<>#&:;%^_~+])!sprintf "^%02X", ord($1)!eg;
    $str =~ s!/!_!g;
    $str =~ s!\\!~!g;
    $str =~ s! !+!g;
    return $str;
}

# Same as mangle_string() but the parameter is a Unix file path.
# Thus '/' is passed through untouched.
sub mangle_path($)
{
    my ($str)=@_;
    $str =~ s!([\x01-\x1f\x80-\xff\$"'`<>#&:;%^_~+])!sprintf "^%02X", ord($1)!eg;
    $str =~ s!\\!~!g;
    $str =~ s! !+!g;
    return $str;
}

# Decodes a mangled string.
sub demangle_string($)
{
    my ($str)=@_;
    $str =~ s!_!/!g;
    $str =~ s!~!\\!g;
    $str =~ s!\+! !g;
    $str =~ s![\^]([0-9a-fA-F]{2})!chr(oct("0x$1"))!eg;
    return $str;
}

sub cmdline2argv($)
{
    my @chars=split "",$_[0];
    my $len=@chars;

    my @argv;
    my $in_quotes=0;
    my $bcount=0;
    my $i=0;
    my $arg="";
    while ($i<$len)
    {
        my $c=$chars[$i++];
        if (($c eq " " or $c eq "\t") and !$in_quotes)
        {
            # Close the argument and copy it
            push @argv,$arg;

            # skip the remaining spaces
            while ($i<$len)
            {
                $c=$chars[$i];
                last if ($c ne " " and $c ne "\t");
                $i++;
            }

            # Start with a new argument
            $arg="";
            $bcount=0;
        }
        elsif ($c eq "\\")
        {
            # '\\'
            $arg.=$c;
            $bcount++;
        }
        elsif ($c eq "\"")
        {
            # '"'
            if (($bcount & 1)==0)
            {
                # Preceded by an even number of '\', this is half that
                # number of '\', plus a quote which we erase.
                my $bq="\\" x ($bcount/2);
                $arg=~s/\\*$/$bq/;
                $in_quotes=!$in_quotes;
            }
            else
            {
                # Preceded by an odd number of '\', this is half that
                # number of '\' followed by a '"'
                my $bq="\\" x ($bcount/2);
                $arg=~s/\\*$/$bq\"/;
            }
            $bcount=0;
        }
        else
        {
            # a regular character
            $arg.=$c;
            $bcount=0;
        }
    }
    push @argv,$arg if ($arg ne "");
    return @argv;
}


#####
#
# Glob functions
#
#####

sub glob2regexp($;$)
{
    my ($glob, $nocase)=@_;
    # Convert the shell glob to a Perl pattern
    $glob =~ s/([].+|^\$\\\(\){}[])/\\$1/g;
    $glob =~ s/\*/.*/g;
    $glob =~ s/^\.\*/[^.].*/;
    $glob =~ s/\?/./g;
    $glob = "(?i:$glob)" if ($nocase);
    return $glob;
}

sub cxglob($$;$)
{
    my ($dir, $globs, $options)=@_;
    $options ||= "";
    my $nocase;
    $nocase=1 if ($options =~ /i/);

    my $matches=[ $dir ];
    foreach my $glob (split m!/!, $globs)
    {
        next if ($glob eq "");

        my $dirs=[];
        if ($nocase or $glob =~ /[?*]/)
        {
            $glob=glob2regexp($glob, $nocase);
            foreach my $dir (@$matches)
            {
                if (opendir(my $dh, $dir))
                {
                    foreach my $dentry (readdir $dh)
                    {
                        push @$dirs, "$dir/$dentry" if ($dentry =~ /^$glob$/);
                    }
                    closedir($dh);
                }
            }
        }
        else
        {
            foreach my $dir (@$matches)
            {
                push @$dirs, "$dir/$glob" if (-l "$dir/$glob" or -e _);
            }
        }
        return () if (!@$dirs);
        $matches=$dirs;
    }
    return @$matches;
}

sub delete_files($$)
{
    my ($dir, $pattern)=@_;

    my $dh;
    if (!opendir($dh, $dir))
    {
        cxlog("unable to open the '$dir' directory: $!\n");
        return 0;
    }
    my $found=0;
    foreach my $dentry (readdir $dh)
    {
        next if ($dentry !~ /$pattern/); # Must filter out '.' and '..' too
        $dentry="$dir/$dentry";
        # $dentry may be a dangling symbolic link, so no -f test
        cxlog("Deleting '$dentry'\n");
        $found=1;
        if (!unlink $dentry)
        {
            cxerr("unable to delete '$dentry': $!\n");
        }
    }
    closedir($dh);
    return $found;
}

sub cxfind($$;$)
{
    my ($rootdir, $pattern, $max)=@_;

    my @matches;
    my @dirs=($rootdir);
    while (@dirs)
    {
        my $dir=shift @dirs;
        if (opendir(my $dh, $dir))
        {
            foreach my $dentry (readdir $dh)
            {
                next if ($dentry =~ /^\.\.?$/);
                if ($dentry =~ /$pattern/)
                {
                    push @matches, "$dir/$dentry";
                    if (defined $max and @matches >= $max)
                    {
                        closedir($dh);
                        return @matches;
                    }
                }
                elsif (!-l "$dir/$dentry" and -d _)
                {
                    push @dirs, "$dir/$dentry";
                }
            }
            closedir($dh);
        }
    }
    return @matches;
}


#####
#
# Miscellaneous
#
#####

sub get_wine_dir_id($)
{
    my ($dev,$ino)=(stat($_[0]))[0,1];

    my $id;
    if ($dev > 0xffffffff)
    {
        $id=sprintf("%lx%08lx-", ($dev >> 32), ($dev & 0xffffffff));
    }
    else
    {
        $id=sprintf("%lx-", $dev);
    }

    if ($ino > 0xffffffff)
    {
        $id.=sprintf("%lx%08lx", ($ino >> 32), ($ino & 0xffffffff));
    }
    else
    {
        $id.=sprintf("%lx", $ino);
    }
    return $id;
}

sub get_unique_id($)
{
    my $id;

    my $uuidgen=cxwhich($ENV{PATH},"uuidgen");
    if (defined $uuidgen)
    {
        $uuidgen=shquote_string($uuidgen);
        $id=`$uuidgen`;
        chomp $id;
    }
    if (!defined $id or $id eq "")
    {
        $id=get_wine_dir_id($_[0]) . "-" . time();
    }
    return $id;
}

# Fast dirname() implementation
sub cxdirname($)
{
    my ($path)=@_;
    return undef if (!defined $path);
    return "." if ($path !~ s!/+[^/]+/*$!!s);
    return "/" if ($path eq "");
    return $path;
}

# Fast basename() implementation
sub cxbasename($)
{
    my ($path)=@_;
    return undef if (!defined $path);
    return "" if ($path =~ s!/+$!!s);
    $path =~ s!^.*/+!!s;
    return $path;
}

# Return 1 if the two paths refer to the same inode, 0 if not
sub same_inode($$)
{
    my ($path1, $path2)=@_;
    return 1 if ($path1 eq $path2);

    my ($dev1, $ino1)=(stat($path1))[0,1];
    my ($dev2, $ino2)=(stat($path2))[0,1];
    return 0 if (!defined $dev1 or !defined $dev2);
    return 1 if ($dev1 == $dev2 and $ino1 == $ino2);
    return 0;
}

sub get_symlink_target($)
{
    my ($link)=@_;
    my $target=readlink($link);
    return $target if ($target =~ m%^/%);
    return cxdirname($link) . "/$target";
}

sub dereference_symlink($)
{
    my ($link)=@_;
    while (-l $link)
    {
        $link=get_symlink_target($link);
        last if (!-e $link);
    }
    return $link;
}

sub cxrealpath($)
{
    my ($path)=@_;

    if ($path !~ m+^/+)
    {
        require Cwd;
        $path=Cwd::cwd() . "/$path";
    }
    my $realpath="";
    foreach my $item (split m%/+%, $path)
    {
        if ($item eq "" or $item eq ".")
        {
            next;
        }
        elsif ($item eq "..")
        {
            if ($realpath eq "")
            {
                # Nothing to do
                ;
            }
            elsif (-d $realpath)
            {
                $realpath=cxdirname($realpath);
            }
            else
            {
                $realpath="$realpath/..";
            }
        }
        else
        {
            $realpath=dereference_symlink("$realpath/$item");
        }
    }
    return $realpath || "/";
}

sub cxmkpath($;$)
{
    my ($path, $mode)=@_;
    # Prevent mkpath from killing the process if it fails!
    require File::Path;
    eval { File::Path::mkpath($path, 0, $mode) };
    return ($@ ? 0 : 1);
}

sub file_grep($$)
{
    my ($filename, $regexp)=@_;

    my $fh;
    if (!open($fh, "<", $filename))
    {
        cxlog("unable to open '$filename' for reading: $!\n");
        return 0;
    }
    while (my $line=<$fh>)
    {
        if ($line =~ /$regexp/)
        {
            close($fh);
            return 1;
        }
    }
    close($fh);
    return 0;
}

sub get_tar()
{
    return cxwhich("$ENV{PATH}:/opt/sfw/bin:/opt/csw/bin", "gtar") || "tar";
}

sub generate_from_template($$$)
{
    my ($dst, $template, $substitutions)=@_;

    my $in;
    if (!open($in, "<", $template))
    {
        cxerr("unable to open '$template' for reading: $!\n");
        return 1;
    }

    my $out;
    if (!open($out, ">", $dst))
    {
        cxerr("unable to open '$dst' for writing: $!\n");
        close($in);
        return 1;
    }

    while (my $line=<$in>)
    {
        if ($line =~ /\@/)
        {
            while (my ($key, $value)=each %{$substitutions})
            {
                cxerr("no value for '$key'\n") if (!defined $value);
                $line =~ s/\@$key\@/$value/g;
            }
            cxwarn("leftover template variable in:\n$line") if ($line =~ /\@[a-z_]+\@/);
        }
        print $out $line;
    }
    close($in);
    close($out);
    return 0;
}

sub garbage_collect_subdirs($$$)
{
    my ($root, $subdir, $delete_root)=@_;
    $subdir="/$subdir" if ($subdir !~ s%^/+%/%);
    while (1)
    {
        if ($subdir eq "/")
        {
            if ($delete_root and rmdir $root)
            {
                cxlog("Deleted '$root'\n");
            }
            last;
        }
        last if (!rmdir "$root$subdir");
        cxlog("Deleted '$root$subdir'\n");
        $subdir=cxdirname($subdir);
    }
}

# Securely take a lock
sub cxlock($$\$)
{
    my ($dir, $file)=@_;

    my @st=lstat($dir);
    if (!@st)
    {
        mkdir($dir, 0700);
        @st=lstat($dir);
        if (!@st)
        {
            $_[2]="unable to create the '$dir' directory: $!";
            return undef;
        }
    }
    if (!-d _)
    {
        $_[2]="'$dir' is not a directory";
        return undef;
    }
    elsif ($st[4] != $>)
    {
        $_[2]="'$dir' is not owned by you";
        return undef;
    }
    elsif ($st[2] & 077)
    {
        $_[2]="'$dir' must not be accessible by other users";
        return undef;
    }
    if (-e "$dir/$file" and (-l "$dir/$file" or !-f "$dir/$file"))
    {
        $_[2]="'$dir/$file' is not a regular file";
        return undef;
    }
    my $lock;
    if (!open($lock, ">", "$dir/$file"))
    {
        $_[2]="unable to create the lock '$dir/$file': $!";
        return undef;
    }
    cxlog("$$: Grabbing the lock '$dir/$file'\n");
    if (!flock($lock, 2))
    {
        $_[2]="unable to lock '$dir/$file': $!";
        close($lock);
        return undef;
    }
    cxlog("$$: Got the lock\n");
    return $lock;
}

sub cxunlock($)
{
    my ($lock)=@_;
    my $rc=1;
    cxlog("$$: Releasing the lock\n");
    $rc=0 if (!flock($lock, 8));
    close($lock);
    return $rc;
}

sub cxexec(@)
{
    cxlog("Exec-ing '", join("' '", @_), "'\n");
    cxlog("-> failed: $!\n") if (!exec @_);
}

sub cxsystem(@)
{
    cxlog("Running '",join("' '",@_),"'\n");
    my $start=CXLog::cxtime();
    my $rc=system(@_);
    cxlog("-> rc=$rc  (took ", CXLog::cxtime()-$start, " seconds)\n");
    return $rc;
}

sub cxbackquote($;$)
{
    my ($cmd, $nolog)=@_;
    cxlog("Running `$cmd`\n");
    my $start=CXLog::cxtime();
    my @output=`$cmd`;
    cxlog("-> rc=$?  (took ", CXLog::cxtime()-$start, " seconds)\n");
    cxlog("output=[@output]\n") if ($? or !$nolog);
    return wantarray ? @output : join("", @output);
}

sub cxmessage(@)
{
    return cxsystem("$ENV{CX_ROOT}/bin/cxmessage",@_);
}

sub cxsu(@)
{
    return cxsystem("$ENV{CX_ROOT}/bin/cxsu",@_);
}


#####
#
# Icon format conversion
#
#####

my %converted;
my $ctp_convert;
my $ctp_xpmtoppm;
my $ctp_pnmtopng;
sub convert_to_png($$;$)
{
    my ($src, $dst, $strict)=@_;
    $dst ||=  "";
    return 1 if ($converted{"$src:$dst"});

    my $png=$dst;
    if (!$png)
    {
        $png=$src;
        $png=~s/\.xpm$/.png/;
    }
    if (-f $src and (!-e $png or (!$strict and -M $src < -M $png) or ($strict and -M $src != -M $png)))
    {
        if (!defined $ctp_convert)
        {
            $ctp_convert=cxwhich($ENV{PATH}, "convert") || "";
            cxlog("ctp_convert='$ctp_convert'\n");
        }
        if (!$ctp_convert or cxsystem($ctp_convert, $src, $png))
        {
            if (!defined $ctp_xpmtoppm)
            {
                $ctp_xpmtoppm=cxwhich($ENV{PATH}, "xpmtoppm") || "";
                $ctp_pnmtopng=cxwhich($ENV{PATH}, "pnmtopng") || "";
                $ctp_xpmtoppm=shquote_string($ctp_xpmtoppm) if ($ctp_xpmtoppm);
                $ctp_pnmtopng=shquote_string($ctp_pnmtopng) if ($ctp_pnmtopng);
                cxlog("ctp_xpmtoppm=$ctp_xpmtoppm ctp_pnmtopng=$ctp_pnmtopng\n");
            }
            if ($ctp_xpmtoppm and $ctp_pnmtopng)
            {
                my $rc;
                my $pbm=shquote_string("$src.pbm");
                my $ppm=shquote_string("$src.ppm");
                my $cmd="$ctp_xpmtoppm --alphaout=$pbm " . shquote_string($src) .
                    " >$ppm";
                if (!cxsystem($cmd))
                {
                    $cmd="$ctp_pnmtopng -alpha $pbm $ppm >" . shquote_string($png);
                    $cmd.=" 2>/dev/null" if (!CXLog::is_on());
                    $rc=1 if (!cxsystem($cmd));
                }
                unlink "$src.pbm";
                unlink "$src.ppm";
                return 0 if (!$rc);
            }
        }
    }
    $converted{"$src:$dst"}=1;
    if ($strict)
    {
        # If looking for an exact timestamp match,
        # copy the last modification time.
        my @st=stat($src);
        utime time(), $st[9], $dst;
    }
    return 1;
}


#####
#
# Character set conversion
#
#####

my $system_encoding;
sub get_system_encoding(;$)
{
    my ($raw)=@_;
    if (!defined $system_encoding)
    {
        $system_encoding = eval "require I18N::Langinfo; I18N::Langinfo->import(qw(langinfo CODESET)); langinfo(CODESET());";
        if ($@ or !$system_encoding)
        {
            $system_encoding=cxbackquote("locale charmap 2>/dev/null");
            if ($? == 0)
            {
                chomp $system_encoding;
            }
            else
            {
                $system_encoding=undef;
            }
        }
        $system_encoding="ANSI_X3.4-1968" if (!$system_encoding);
        cxlog("system encoding='$system_encoding'\n");
    }
    my $encoding=$system_encoding;
    if (!$raw and $encoding eq "ANSI_X3.4-1968")
    {
        cxlog("overriding '$encoding' system encoding\n");
        $encoding="ISO-8859-1";
    }
    return $encoding;
}


#####
#
# Localization
#
#####

my $cxlocales;
sub get_supported_locales()
{
    if (!defined $cxlocales and opendir(my $dh, "$ENV{CX_ROOT}/share/locale"))
    {
        foreach my $dentry (readdir $dh)
        {
            push @$cxlocales, $dentry if ($dentry =~ /^[a-z]{2}/);
        }
        closedir($dh);
    }
    return @$cxlocales;
}

my $cxlang="";
sub cxgetlang()
{
    return $cxlang;
}

sub cxsetlang($)
{
    my $oldlang=$cxlang;
    $cxlang=$_[0] || "";
    return $oldlang;
}
my $cxencoding="";
sub cxgetencoding()
{
    return $cxencoding;
}

sub cxsetencoding($)
{
    my $oldencoding=$cxencoding;
    $cxencoding=$_[0] || "";
    return $oldencoding;
}

my %cxgettext_cache;
my $_non_c_locale;
sub cxgettext(@)
{
    my $format=shift @_;
    my $key="$cxlang:$cxencoding:$format";
    my $local_format=$cxgettext_cache{$key};
    if (!defined $local_format)
    {
        my $cmd="";
        if ($cxlang)
        {
            if (!defined $_non_c_locale)
            {
                $_non_c_locale=cxbackquote("locale -a | egrep -v '^(C|POSIX)\$' | head -n 1");
                chomp $_non_c_locale;
            }
            $cmd.="LC_ALL=$_non_c_locale LANGUAGE=$cxlang ";
        }
        $cmd.=shquote_string("$ENV{CX_ROOT}/bin/cxgettext") .
            " --textdomain crossover --textdomaindir " .
            shquote_string("$ENV{CX_ROOT}/share/locale") . " ";
        $cmd.="--encoding $cxencoding " if ($cxencoding);
        $cmd.=shquote_string($format);
        $cmd.=" 2>/dev/null" if (!CXLog::is_on());
        $local_format=cxbackquote($cmd);
        $local_format=$format if ($? != 0 or $local_format eq "");
        $cxgettext_cache{$key}=$local_format;
    }
    return sprintf($local_format,@_);
}


#####
#
# fstab parsing
#
#####

sub parse_fstab_get_fstypes($$)
{
    my ($fstype, $options)=@_;

    # fstype can be a list of filesystem types
    # so normalize it to make it easy to use regexps
    $fstype =~ s/^,*/,/;
    $fstype =~ s/,*$/,/;
    $fstype =~ s/,+/,/g;

    my $rfstype;
    if ($fstype =~ /^,(?:subfs|supermount),$/)
    {
        # supermount and submount use the fs option to specify the
        # filesystem type. For submount, it can even be a colon-separated
        # list.
        $rfstype=(grep /^fs=/, @$options)[0];
        if ($rfstype)
        {
            $rfstype =~ s/^fs=:*/,/;
            $rfstype =~ s/:*$/,/;
            $rfstype =~ s/:+/,/g;
            # submount's cdfss is synonymous for 'iso9660 or udf'
            $rfstype =~ s/,cdfss,/,iso9660,udf,/;
        }
    }
    $rfstype||=$fstype;
    return ($fstype, $rfstype);
}

sub parse_fstab_get_device($$$)
{
    my ($device, $nfstype, $options)=@_;

    if ($nfstype eq ",supermount,")
    {
        # Supermount specifies the device in the options
        my $opt=(grep /^dev=/, @$options)[0];
        if (defined $opt)
        {
            $device=$opt;
            $device =~ s/^dev=//;
        }
    }
    return $device;
}

sub parse_fstab($)
{
    my ($fstabfile)=@_;

    my $fh;
    if (!open($fh, "<", $fstabfile))
    {
        cxlog("unable to open '$fstabfile' for reading: $!\n");
        return ([], []);
    }

    my ($floppies, $cdroms, %mountpoints);
    foreach my $line (<$fh>)
    {
        cxlog($line);
        # Ignore empty lines and comments
        $line =~ s/#.*$//;
        next if ($line =~ /^\s*$/);
        chomp $line;

        my @fields = split /\s+/, $line;
        if (@fields < 4)
        {
            cxlog(" -> skipped, not enough fields\n");
            next;
        }

        # Skip duplicate entries
        my $mpoint=$fields[1];
        next if ($mountpoints{$mpoint});
        $mountpoints{$mpoint}=1;

        # Extract the basic data
        my @options=split /,/, $fields[3];
        my ($nfstype, $rfstype)=CXUtils::parse_fstab_get_fstypes($fields[2], \@options);
        my $device=CXUtils::parse_fstab_get_device($fields[0], $nfstype, \@options);

        my $has_user=(grep /^user$/, @options) ? 1 : 0;
        my $has_unhide=(grep /^unhide$/, @options) ? 1 : 0;
        my $has_loop=(grep /^loop$/, @options) ? 1 : 0;
        my $descriptor=[$mpoint, $device, $has_user, $has_unhide, $has_loop];

        if ($rfstype =~ /,(?:iso9660|udf),/ or
            ($rfstype =~ /,auto,/ and
             ($device =~ /(?:cdr|dvd)/ or $mpoint =~ /(?:cdr|dvd)/)
            )
           )
        {
            push @$cdroms, $descriptor;
        }
        elsif ($rfstype =~ /,(?:v?fat|floppyfss),/ and
               ($device =~ m!^/dev/fd! or $mpoint =~ /floppy/)
              )
        {
            push @$floppies, $descriptor;
        }
    }
    close($fh);

    return ($cdroms, $floppies);
}

sub parse_automaster($)
{
    my ($automasterfile)=@_;
    my @floppies = ();
    my @cdroms = ();
    my @mountpoints = ();

    my $masterfh;
    if (!open($masterfh, "<", $automasterfile))
    {
        cxlog("unable to open '$automasterfile' for reading: $!\n");
        return ([@cdroms], [@floppies]);
    }

    foreach my $line (<$masterfh>)
    {
        # Ignore commented lines
        next if ($line =~ /^\s*\#/);
        next if ($line =~ /^\+/); # Ignore include directives
        next if ($line =~ /^\s*$/);

        my $entry = $line;
        cxlog($entry);
        $entry =~ s/(\#.*)$//;
        my @fields = split /\s+/, $entry;
        if (@fields == 1)
        {
            cxlog("ignoring unsupported '$automasterfile' line format '$entry'\n");
            next;
        }

        # Parse line
        my $automountpoint=$fields[0];
        my @map=split /:/, $fields[1];

        my ($maptype, $mapname);
        if (@map == 1)
        {
            $mapname=$map[0];
            if ($mapname !~ m%^/% and -e "/etc/$mapname")
            {
                $mapname="/etc/$mapname";
            }
            $maptype=(-x $mapname ? "program" : "file");
        }
        else
        {
            $maptype=$map[0];
            $mapname=$map[1];
        }

        if ($maptype ne "file")
        {
            cxlog("unsupported map type '$maptype'. Skipping '$automountpoint'\n");
            next;
        }

        my $mapfh;
        if (!open($mapfh, "<", $mapname))
        {
            cxlog("unable to open '$mapname' for reading: $!\n");
            next;
        }

        my (@tmpfloppies, @tmpcdroms);
        foreach my $mapline (<$mapfh>)
        {
            # Ignore commented lines
            next if ($mapline =~ /^\s*\#/);
            next if ($mapline =~ /^\s*$/);

            my @maplinefields = split /\s+/, $mapline;
            next if (@maplinefields < 2);

            my ($location, $key);
            if (@maplinefields == 2)
            {
                # No options apparently, the location is in the second field
                $key=$maplinefields[0];
                $location=$maplinefields[1];
            }
            else
            {
                # The options are in the second field, the location is 
                # in the third!
                $key=$maplinefields[0];
                $location=$maplinefields[2];
            }

            if ($location !~ /^:/)
            {
                # This is either an NFS mount or a multi-mount
                # and we don't care about those
                next;
            }

            my $device = $location;
            $device =~ s/^://;

            my $thismountpoint="$automountpoint/$key";
            # Ignore double entries
            if ((grep /$thismountpoint/,@mountpoints)[0])
            {
                cxlog("Duplicate entry for '$thismountpoint'. Skipping it\n");
                next;
            }
            push @mountpoints, $thismountpoint;

            # Handle floppies
            if ($mapline =~ /floppy/ or $mapline =~ /\/dev\/fd/)
            {
                push @floppies, [$thismountpoint, $device];
            }

            if ($mapline =~ /iso9660/ or $mapline =~ /cdrom/ or
                $mapline =~ /\/dvd/ or $mapline =~ /\/cdr/)
            {
                push @cdroms, [$thismountpoint, $device];
            }
        }
        close($mapfh);
    }
    close($masterfh);

    return ([@cdroms], [@floppies]);
}

return 1;
